from sunpy.data.test import get_test_filepath

MEDN_IMAGE = get_test_filepath('waveunit/medn_halph_fl_20050501_074655.header')
MQ_IMAGE = get_test_filepath('waveunit/mq130812.084253.header')
NA_IMAGE = get_test_filepath('waveunit/na120701.091058.header')
SVSM_IMAGE = get_test_filepath('waveunit/svsm_e3100_S2_20110625_1856.header')
