"""
This subpackage contains a series of file readers and only for internal use.

None of this code is intended to be used directly by a user and is not part of the sunpy public API
and is not added to the API reference pages.
"""
